'use strict';

const yaml = require('js-yaml');
const fs = require('fs');
const to = require('await-to-js').to;
const Router = require('restify-router').Router;

class RouteParser {
  constructor(options) {
    this.server = options.server;
    this.appRoot = options.appRoot || './';
    this.filePath = options.filePath || 'src/constants/routes.yml';
    this.providersPath = options.providersPath || 'providers';

    this.router = new Router();
  }

  parse() {
    this.getJSON();
    this.validateConfig();
    this.configureRoutes();
  }

  getJSON() {
    let content;
    let routesConfig;

    content = fs.readFileSync(this.filePath, 'utf8');
    if (this.filePath.indexOf('.yml')) {
      routesConfig = yaml.load(content);
    } else {
      routesConfig = JSON.parse(content);
    }

    this.routes = routesConfig.routes;
  }

  validateConfig() {
    const routes = this.routes;
    const err = RouteParser.throwError;
    if (!Array.isArray(routes))
      err('routes field should contain an array', routes);
    for (let i = 0, len = routes.length; i < len; ++i) {
      const route = routes[i];
      if (typeof route !== 'object')
        err('route should be an object', route);
      if (typeof route.prefix !== 'string')
        err('route.prefix should be a string', route.prefix);
      if (typeof route.controller !== 'string' && typeof route.controller !== 'object')
        err('route.controller should be a string or an object', route.controller);

      const endpoints = route.endpoints;
      if (!Array.isArray(endpoints))
        err('route.endpoints should be an array', endpoints);

      for (let j = 0; j < endpoints.length; ++j) {
        const endpoint = endpoints[j];
        if (typeof endpoint !== 'object')
          err('route.endpoints.endpoint should be an object', endpoint);
        if (typeof endpoint.route !== 'string')
          err('route.endpoints.endpoint.route should be a string', endpoint.route);
        // TODO validate the rest
      }
    }
  }

  static throwError(message, currentValue) {
    throw new Error(`RouteParser error: ${message}. Current value: ${currentValue}`);
  }

  configureRoutes() {
    const routes = this.routes;
    for (let i = 0, len = routes.length; i < len; ++i) {
      const route = routes[i];
      const routers = this.applyEndpoints(route);
      this.router.add(route.prefix, routers);
    }
    this.router.applyRoutes(this.server);
  }

  applyEndpoints(route) {
    const endpoints = route.endpoints;
    let Controller;

    // import controller from project or from package
    if (typeof route.controller === 'object') {
      const controllerName = route.controller.name;
      if (route.controller.external) {
        Controller = require(controllerName);
      } else {
        route.controller = controllerName;
      }
    }
    if (typeof route.controller === 'string') {
      let separatedControllerPath = route.controller.split('/');
      const controllerName = separatedControllerPath[separatedControllerPath.length - 1];
      Controller = require(
        `${this.appRoot}/${this.providersPath}/${route.controller}/` +
        `${controllerName}.controller.js`
      );
    }
    const controller = new Controller();

    const router = new Router();

    for (let i = 0, len = endpoints.length; i < len; ++i) {
      const endpoint = endpoints[i];
      const routeInfo = endpoint.route.split(' ');
      let method, path;
      if (routeInfo.length === 2) {
        method = routeInfo[0].toLowerCase();
        path = routeInfo[1];
      } else if (routeInfo.length === 1) {
        method = endpoint.method.toLowerCase();
        path = endpoint.route;
      }

      const routeMethod = async (req, res, next) => {
        const requiredFields = endpoint.required;
        if (!req.body && requiredFields.length) {
          res.send(400, {
            code: 'missing_fields',
            message: 'No parameters provided.',
          });
        }
        const params = req.body;
        const fields = Object.keys(params);
        let missingFields = [];
        for (let i = 0, len = requiredFields.length; i < len; ++i) {
          const requiredField = requiredFields[i];
          if (!fields.includes(requiredField.name)) {
            missingFields.push(requiredField);
          }
        }
        if (missingFields.length) {
          res.send(400, {
            code: 'missing_fields',
            message: `Missing required parameters: ${missingFields.join(', ')}`,
          });
        } else {
          let err, data;
          [err, data] = await to(controller[endpoint.handler](params));
          if (err) {
            const statusCode = err.statusCode;
            delete err.statusCode;
            res.send(statusCode, err);
          } else {
            const statusCode = data.statusCode;
            delete data.statusCode;
            res.send(statusCode, data);
          }
        }
        next();
      };

      if (endpoint.secured || route.secured) {
        // TODO authenticate
        router[method](path, routeMethod);
      } else {
        router[method](path, routeMethod);
      }
    }

    return router;
  }
}

module.exports = RouteParser;
